# -*- coding: utf-8 -*-

"""Mixin class and utilities for handling ReaxFF forcefields"""

import argparse
from datetime import date
import logging
import packaging.version
from pathlib import Path
import pprint  # noqa: F401
import sys

from tabulate import tabulate

from .metadata import metadata as ff_metadata
from seamm_util import element_data

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")


metadata = {
    "general parameters": [
        {
            "id": 1,
            "description": "Overcoordination parameter 1",
            "parameter": "Pboc,1",
        },
        {
            "id": 2,
            "description": "Overcoordination parameter 2",
            "parameter": "Pboc,2",
        },
        {
            "id": 3,
            "description": "Valency angle conjugation parameter",
            "parameter": "Pcoa,2",
        },
        {
            "id": 4,
            "description": "Triple bond stabilisation parameter",
            "parameter": "Ptrip,4",
        },
        {
            "id": 5,
            "description": "Triple bond stabilisation parameter",
            "parameter": "Ptrip,3",
        },
        {
            "id": 6,
            "description": "C2-correction",
            "parameter": "kc2",
        },
        {
            "id": 7,
            "description": "Over/undercoordination parameter 6",
            "parameter": "Povun,6",
        },
        {
            "id": 8,
            "description": "Triple bond stabilisation parameter 2",
            "parameter": "Ptrip,2",
        },
        {
            "id": 9,
            "description": "Over/undercoordination parameter 7",
            "parameter": "Povun,7",
        },
        {
            "id": 10,
            "description": "Over/undercoordination parameter 8",
            "parameter": "Povun,8",
        },
        {
            "id": 11,
            "description": "Triple bond stabilization parameter 1",
            "parameter": "Ptrip,1",
        },
        {
            "id": 12,
            "description": "Lower Taper-radius",
            "parameter": "Rtaper,lower",
        },
        {
            "id": 13,
            "description": "Upper Taper-radius",
            "parameter": "Rtaper,upper",
        },
        {
            "id": 14,
            "description": "Fe dimer correction?",
            "parameter": "Pfe1",
        },
        {
            "id": 15,
            "description": "Valency undercoordination",
            "parameter": "Pval,7",
        },
        {
            "id": 16,
            "description": "Valency angle/lone pair parameter",
            "parameter": "Plp,1",
        },
        {
            "id": 17,
            "description": "Valency angle",
            "parameter": "Pval,9",
        },
        {
            "id": 18,
            "description": "Valency angle parameter",
            "parameter": "Pval,10",
        },
        {
            "id": 19,
            "description": "Not used",
            "parameter": "not_used_1",
        },
        {
            "id": 20,
            "description": "Double bond/angle parameter",
            "parameter": "Ppen,2",
        },
        {
            "id": 21,
            "description": "Double bond/angle parameter: overcoord",
            "parameter": "Ppen,3",
        },
        {
            "id": 22,
            "description": "Double bond/angle parameter: overcoord",
            "parameter": "Ppen,4",
        },
        {
            "id": 23,
            "description": "Not used",
            "parameter": "not_used_2",
        },
        {
            "id": 24,
            "description": "Torsion/BO parameter",
            "parameter": "Ptor,2",
        },
        {
            "id": 25,
            "description": "Torsion overcoordination",
            "parameter": "Ptor,3",
        },
        {
            "id": 26,
            "description": "Torsion overcoordination",
            "parameter": "Ptor,4",
        },
        {
            "id": 27,
            "description": "Conjugation 0 (not used)",
            "parameter": "not_used_3",
        },
        {
            "id": 28,
            "description": "Conjugation",
            "parameter": "Pcot,2",
        },
        {
            "id": 29,
            "description": "vdWaals shielding",
            "parameter": "PvdW,1",
        },
        {
            "id": 30,
            "description": "Cutoff for bond order (*100)",
            "parameter": "BO_cutoff",
        },
        {
            "id": 31,
            "description": "Valency angle conjugation parameter",
            "parameter": "Pcoa,4",
        },
        {
            "id": 32,
            "description": "Overcoordination parameter",
            "parameter": "Povun,4",
        },
        {
            "id": 33,
            "description": "Overcoordination parameter",
            "parameter": "Povun,3",
        },
        {
            "id": 34,
            "description": "Valency/lone pair parameter",
            "parameter": "Pval,8",
        },
        {
            "id": 35,
            "description": "Not used",
            "parameter": "not_used_4",
        },
        {
            "id": 36,
            "description": "Not used",
            "parameter": "not_used_5",
        },
        {
            "id": 37,
            "description": "Molecular energy (not used)",
            "parameter": "not_used_6",
        },
        {
            "id": 38,
            "description": "Molecular energy (not used)",
            "parameter": "not_used_7",
        },
        {
            "id": 39,
            "description": "Valency angle conjugation parameter",
            "parameter": "Pcoa,3",
        },
    ],
    "atomic parameters": [
        {
            "id": 1,
            "description": "Sigma bond covalent radius",
            "parameter": "R0,alpha",
            "units": "Å",
        },
        {
            "id": 2,
            "description": "Valency",
            "parameter": "Val",
        },
        {
            "id": 3,
            "description": "atomic mass",
            "parameter": "m",
            "units": "Dalton",
        },
        {
            "id": 4,
            "description": "van der Waals radius",
            "parameter": "RvdW",
            "units": "Å",
        },
        {
            "id": 5,
            "description": "van der Waals well depth",
            "parameter": "Dij",
            "units": "kcal/mol",
        },
        {
            "id": 6,
            "description": "gamma for electron equilibration method (EEM)",
            "parameter": "gamma",
            "units": "1/Å",
        },
        {
            "id": 7,
            "description": "Pi bond covalent radius",
            "parameter": "R0,pi",
            "units": "Å",
        },
        {
            "id": 8,
            "description": "number of valence electrons",
            "parameter": "Val,e",
        },
        {
            "id": 9,
            "description": "van der Waals parameter",
            "parameter": "alpha",
        },
        {
            "id": 10,
            "description": "van der Waals shielding",
            "parameter": "gamma,w",
            "units": "1/Å",
        },
        {
            "id": 11,
            "description": "valency for 1-3 bond order correction",
            "parameter": "Val,angle",
        },
        {
            "id": 12,
            "description": "Over/undercoordination parameter 5",
            "parameter": "Povun,5",
            "units": "kcal/mol",
        },
        {
            "id": 13,
            "description": "eReaxff atom type parameter. LAMMPS does not use.",
            "parameter": "not_used_1",
        },
        {
            "id": 14,
            "description": "electronegativity for electron equilibration method (EEM)",
            "parameter": "chi",
            "units": "eV",
        },
        {
            "id": 15,
            "description": "hardness for electron equilibration method (EEM)",
            "parameter": "eta",
            "units": "eV",
        },
        {
            "id": 16,
            "description": "Donor or acceptor switch in h-bonds, integer",
            "parameter": "Phbond",
        },
        {
            "id": 17,
            "description": "bond covalent radius with two pi orbitals",
            "parameter": "R0,pi-pi",
        },
        {
            "id": 18,
            "description": "Energy factor of lone pairs",
            "parameter": "Plp,2",
            "units": "kcal/mol",
        },
        {
            "id": 19,
            "description": "atomic enthalpy of formation. LAMMPS does not use.",
            "parameter": "Hat",
        },
        {
            "id": 20,
            "description": "bond order correction 4",
            "parameter": "Pboc,4",
        },
        {
            "id": 21,
            "description": "bond order correction 3",
            "parameter": "Pboc,3",
        },
        {
            "id": 22,
            "description": "bond order correction 5",
            "parameter": "Pboc,5",
        },
        {
            "id": 23,
            "description": "atomic softness cutoff parameter",
            "parameter": "C_i",
        },
        {
            "id": 24,
            "description": "eReaxFF constant depending on atom. LAMMPS does not use.",
            "parameter": "alpha_e",
        },
        {
            "id": 25,
            "description": "Over/undercoordination parameter 2 for valence angle",
            "parameter": "Povun,2",
        },
        {
            "id": 26,
            "description": "valence angle parameter",
            "parameter": "Pval,3",
        },
        {
            "id": 27,
            "description": "eReaxFF constant depending on atom. LAMMPS does not use.",
            "parameter": "beta",
        },
        {
            "id": 28,
            "description": "number of lone pairs",
            "parameter": "Val,boc",
        },
        {
            "id": 29,
            "description": "valence angle parameter 5",
            "parameter": "Pval,5",
        },
        {
            "id": 30,
            "description": "inner wall vdW repulsion parameter",
            "parameter": "Rcore,2",
        },
        {
            "id": 31,
            "description": "inner wall vdW repulsion parameter",
            "parameter": "Ecore,2",
        },
        {
            "id": 32,
            "description": "inner wall vdW repulsion parameter",
            "parameter": "Acore,2",
        },
    ],
    "bond parameters": [
        {
            "id": 1,
            "description": "Sigma bond dissociation energy",
            "parameter": "De,sigma",
            "units": "kcal/mol",
        },
        {
            "id": 2,
            "description": "Pi bond dissociation energy",
            "parameter": "De,pi",
            "units": "kcal/mol",
        },
        {
            "id": 3,
            "description": "Double pi bond dissociation energy",
            "parameter": "De,pi-pi",
            "units": "kcal/mol",
        },
        {
            "id": 4,
            "description": "bond energy parameter",
            "parameter": "Pbe,1",
        },
        {
            "id": 5,
            "description": "double pi bond parameter",
            "parameter": "Pbo,5",
        },
        {
            "id": 6,
            "description": "1-3 bond order correction",
            "parameter": "13_boc",
        },
        {
            "id": 7,
            "description": "double pi bond order",
            "parameter": "Pbo,6",
        },
        {
            "id": 8,
            "description": "Overcoordination penalty",
            "parameter": "Povun,1",
        },
        {
            "id": 9,
            "description": "bond energy parameter",
            "parameter": "Pbe,2",
        },
        {
            "id": 10,
            "description": "pi bond order parameter",
            "parameter": "Pbo,3",
        },
        {
            "id": 11,
            "description": "pi bond order parameter",
            "parameter": "Pbo,4",
        },
        {
            "id": 12,
            "description": "not used. LAMMPS does not use.",
            "parameter": "not_used_1",
        },
        {
            "id": 13,
            "description": "sigma bond order",
            "parameter": "Pbo_1",
        },
        {
            "id": 14,
            "description": "sigma bond order",
            "parameter": "Pbo,2",
        },
        {
            "id": 15,
            "description": "uncorrected BO overcoordination",
            "parameter": "ovc",
        },
        {
            "id": 16,
            "description": "LAMMPS does not use, nor read",
            "parameter": "not_used_2",
        },
    ],
    "off-diagonal parameters": [
        {
            "id": 1,
            "description": "van der Waals energy",
            "parameter": "Dij",
            "units": "kcal/mol",
        },
        {
            "id": 2,
            "description": "van der Waals radius",
            "parameter": "RvdW",
            "units": "Å",
        },
        {
            "id": 3,
            "description": "van der Waals parameter",
            "parameter": "alpha",
        },
        {
            "id": 4,
            "description": "sigma bond length",
            "parameter": "R0,sigma",
            "units": "Å",
        },
        {
            "id": 5,
            "description": "pi bond length",
            "parameter": "R0,pi",
            "units": "Å",
        },
        {
            "id": 6,
            "description": "pi-pi bond length",
            "parameter": "R0,pi-pi",
            "units": "Å",
        },
    ],
    "angle parameters": [
        {
            "id": 1,
            "description": "180° - equilibrium angle",
            "parameter": "Theta0",
            "units": "degree",
        },
        {
            "id": 2,
            "description": "valence angle parameter 1",
            "parameter": "Pval,1",
            "units": "kcal/mol",
        },
        {
            "id": 3,
            "description": "valence angle parameter 2",
            "parameter": "Pval,2",
        },
        {
            "id": 4,
            "description": "valence conjugation",
            "parameter": "Pcoa,1",
            "units": "kcal/mol",
        },
        {
            "id": 5,
            "description": "Undercoordination penalty",
            "parameter": "Pval,7",
        },
        {
            "id": 6,
            "description": "penalty energy",
            "parameter": "Ppen,1",
        },
        {
            "id": 7,
            "description": "valence angle parameter 4",
            "parameter": "Pval,4",
        },
    ],
    "torsion parameters": [
        {
            "id": 1,
            "description": "V1 torsion barrier",
            "parameter": "V1",
            "units": "kcal/mol",
        },
        {
            "id": 2,
            "description": "V2 torsion barrier",
            "parameter": "V2",
            "units": "kcal/mol",
        },
        {
            "id": 3,
            "description": "V3 torsion barrier",
            "parameter": "V3",
            "units": "kcal/mol",
        },
        {
            "id": 4,
            "description": "Torsion angle parameter",
            "parameter": "Ptor,1",
        },
        {
            "id": 5,
            "description": "Conjugation energy",
            "parameter": "Pcot,1",
            "units": "kcal/mol",
        },
        {
            "id": 6,
            "description": "LAMMPS does not use, nor read",
            "parameter": "not_used_1",
        },
        {
            "id": 7,
            "description": "LAMMPS does not use, nor read",
            "parameter": "not_used_2",
        },
    ],
    "hydrogen bond parameters": [
        {
            "id": 1,
            "description": "hydrogen bond length",
            "parameter": "Rhb",
            "units": "Å",
        },
        {
            "id": 2,
            "description": "hydrogen bond energy",
            "parameter": "Ehb",
            "units": "kcal/mol",
        },
        {
            "id": 3,
            "description": "hydrogen bond angle",
            "parameter": "Thb",
            "units": "degree",
        },
        {
            "id": 4,
            "description": "LAMMPS does read",
            "parameter": "Phb3",
        },
    ],
}

# Add to the main forcefield metadata
ff_metadata["reaxff_general_parameters"] = {
    "equation": [],
    "constants": [
        ("Parameter", ""),
        ("Value", ""),
        ("Description", ""),
    ],
    "topology": {
        "type": "reaxff general parameters",
        "n_atoms": 0,
        "symmetry": "none",
        "fill": 0,
        "flip": 0,
    },
}

# The atomic parameter sections
n_ap = len(metadata["atomic parameters"])
tmp = sorted(metadata["atomic parameters"], key=lambda k: k["parameter"])
for g0 in range(0, n_ap, 8):
    g1 = min(g0 + 8, n_ap)
    ff_metadata[f"reaxff_atomic_parameters_{g0 + 1}-{g1}"] = {
        "equation": [],
        "constants": [
            (d["parameter"], d["units"] if "units" in d else "") for d in tmp[g0:g1]
        ],
        "topology": {
            "type": "reaxff atomic parameters",
            "n_atoms": 1,
            "symmetry": "none",
            "fill": 0,
            "flip": 0,
        },
    }

# The bond parameter sections
n_bp = len(metadata["bond parameters"])
tmp = sorted(metadata["bond parameters"], key=lambda k: k["parameter"])
for g0 in range(0, n_bp, 8):
    g1 = min(g0 + 8, n_ap)
    ff_metadata[f"reaxff_bond_parameters_{g0 + 1}-{g1}"] = {
        "equation": [],
        "constants": [
            (d["parameter"], d["units"] if "units" in d else "") for d in tmp[g0:g1]
        ],
        "topology": {
            "type": "reaxff bond parameters",
            "n_atoms": 2,
            "symmetry": "none",
            "fill": 0,
            "flip": 0,
        },
    }

# The off-diagonal parameter sections
tmp = sorted(metadata["off-diagonal parameters"], key=lambda k: k["parameter"])
ff_metadata["reaxff_off-diagonal_parameters"] = {
    "equation": [],
    "constants": [(d["parameter"], d["units"] if "units" in d else "") for d in tmp],
    "topology": {
        "type": "reaxff off-diagonal parameters",
        "n_atoms": 2,
        "symmetry": "none",
        "fill": 0,
        "flip": 0,
    },
}

# The angle parameter sections
tmp = sorted(metadata["angle parameters"], key=lambda k: k["parameter"])
ff_metadata["reaxff_angle_parameters"] = {
    "equation": [],
    "constants": [(d["parameter"], d["units"] if "units" in d else "") for d in tmp],
    "topology": {
        "type": "reaxff angle parameters",
        "n_atoms": 3,
        "symmetry": "none",
        "fill": 0,
        "flip": 0,
    },
}

# The torsion parameter sections
tmp = sorted(metadata["torsion parameters"], key=lambda k: k["parameter"])
ff_metadata["reaxff_torsion_parameters"] = {
    "equation": [],
    "constants": [(d["parameter"], d["units"] if "units" in d else "") for d in tmp],
    "topology": {
        "type": "reaxff torsion parameters",
        "n_atoms": 4,
        "symmetry": "none",
        "fill": 0,
        "flip": 0,
    },
}

# The hydrogen bond parameter sections
tmp = sorted(metadata["hydrogen bond parameters"], key=lambda k: k["parameter"])
ff_metadata["reaxff_hydrogen-bond_parameters"] = {
    "equation": [],
    "constants": [(d["parameter"], d["units"] if "units" in d else "") for d in tmp],
    "topology": {
        "type": "reaxff hydrogen bond parameters",
        "n_atoms": 3,
        "symmetry": "none",
        "fill": 0,
        "flip": 0,
    },
}


def import_reaxff(options):
    """Import a ReaxFF forcefield in the Reax format"""
    reaxff = Path(options.reaxff).expanduser()
    output = Path(options.output).expanduser()

    if options.personal:
        output = Path("~/.seamm.d/data/Forcefields/").expanduser() / output.name

    if "ffname" not in options:
        ffname = output.stem
    else:
        ffname = options.ffname

    version = options.version if "version" in options else None

    print(f"Importing ReaxFF forcefield from {reaxff} to {output}")

    parameters = read_reaxff_parameters(reaxff)

    write_ff_file(output, parameters, ffname=ffname, version=version)


def parse_reax_data(lines):
    """Parse ReaxFF data from a list of lines

    Note
    ----
    It appears that the codes reading the forcefield files are not very fussy. The file
    contains a header line for each section giving the number of atoms or terms in the
    section. However, the code does not appear to check that the number of atoms or
    terms actually match that given.

    Therefore this code will warn about such mismatches, but not raise an exception. It
    appears that the header line separating sections has a sinlge integer, perhaps
    followed by a comment. We will use that as a marker to separate sections.
    """
    parameters = {}
    it = iter(lines)
    parameters["comment"] = next(it)

    # Handle the general parameters
    header = next(it)
    n_general_parameters = int(header.split()[0])
    mdata = metadata["general parameters"]
    gp = parameters["general parameters"] = {}
    for i, line in enumerate(it):
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            nelements = int(line[0])
            break
        gp[mdata[i]["parameter"]] = line[0]

    if i != n_general_parameters:
        logger.warning(
            f"Expected {n_general_parameters} general parameters, but found {i}"
        )

    # Handle the atomic parameters... the section looks like this:
    #
    #  7    ! Nr of atoms; cov.r; valency;a.m;Rvdw;Evdw;gammaEEM;cov.r2;#
    #            alfa;gammavdW;valency;Eunder;Eover;chiEEM;etaEEM;n.u.
    #            cov r3;Elp;Heat inc.;n.u.;n.u.;n.u.;n.u.
    #            ov/un;val1;n.u.;val3,vval4
    # Li   2.1097   1.0000   6.9410   2.1461   0.3726   0.8651  -0.1000   1.0000
    #      9.0000   1.2063   1.0000   0.0000   0.0000  -6.2351  12.7757   0.0000
    #     -1.0000   0.0000  37.5000   5.4409   6.9107   0.1973   0.8563   0.0000
    #    -17.1659   2.2989   1.0338   1.0000   2.8103   1.3000   0.2000  13.0000
    # Si   2.0057   4.0000  28.0600   1.7098   0.2683   0.3041   1.2635   4.0000
    # ...

    n_atoms = 0
    parameters["atomic parameters"] = {}
    mdata = metadata["atomic parameters"]
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        line_sv = line
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            nbonds = int(line[0])
            break
        if len(line) > 1 and line[0].isalpha() and len(line[0]) <= 2:
            # Looks like an element!
            symbol = line[0]
            # Except use X as a placeholder...so just skip
            if symbol == "X":
                next(it)
                next(it)
                next(it)
                continue

            if len(line) != 9:
                raise ValueError(
                    f"Expected 8 element parameters, but found {len(line) - 1}"
                    f"\n\t{line_sv}"
                )
            ap = parameters["atomic parameters"][symbol] = {}
            n_atoms += 1
            count = 0
            for val in line[1:]:
                ap[mdata[count]["parameter"]] = val
                count += 1
            for i in range(3):
                line = next(it).split("!")[0].split()
                if len(line) != 8:
                    raise ValueError(
                        f"Expected 8 element parameters, but found {len(line)}"
                        f"\n\t{line_sv}"
                    )
                for val in line:
                    ap[mdata[count]["parameter"]] = val
                    count += 1
    if nelements != n_atoms:
        logger.warning(f"Expected {nelements} atomic parameters, but found {n_atoms}")

    # Handle the bond parameters... the section looks like this:
    #
    # 20      ! Nr of bonds; Edis1;LPpen;n.u.;pbe1;pbo5;13corr;pbo6
    #                         pbe2;pbo3;pbo4;n.u.;pbo1;pbo2;ovcorr
    #  1  1  59.2876   0.0000   0.0000   0.4363   0.3000   0.0000  26.0000   0.4590
    #         0.0485   0.0000  12.0000   1.0000  -0.1990   4.1914   0.0000   0.0000
    #  2  2  71.2687  45.8533  17.1960  -0.4953  -0.3182   1.0000  15.9405   0.0100
    # ...

    n_bonds = 0
    parameters["bond parameters"] = {}
    mdata = metadata["bond parameters"]
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        line_sv = line
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            noffdiagonals = int(line[0])
            break
        if len(line) > 2 and line[0].isdecimal() and line[1].isdecimal():
            # Looks like a bond!
            if len(line) != 10:
                raise ValueError(
                    f"Expected 8 bond parameters, but found {len(line) - 2}"
                    f"\n\t{line_sv}"
                )
            key = (int(line[0]), int(line[1]))
            n_bonds += 1
            bp = parameters["bond parameters"][key] = {}
            count = 0
            for val in line[2:]:
                bp[mdata[count]["parameter"]] = val
                count += 1
            line = next(it).split("!")[0].split()
            if len(line) != 8:
                raise ValueError(
                    "Expected 8 bond parameters, but found {len(line)}\n\t{line_sv}"
                )
            for val in line:
                bp[mdata[count]["parameter"]] = val
                count += 1
    if nbonds != n_bonds:
        logger.warning(f"Expected {nbonds} bond parameters, but found {n_bonds}")

    # Handle the off-diagonal parameters... the section looks like this:
    #
    # 14    ! Nr of off-diagonal terms; Ediss;Ro;gamma;rsigma;rpi;rpi2
    #  2  1   0.3183   2.4157   9.0000   2.1443   1.0000   1.0000
    #  3  1   0.3984   2.1335   9.6455   1.5037   1.0000   1.0000
    # ...

    n_offdiagonals = 0
    mdata = metadata["off-diagonal parameters"]
    parameters["off-diagonal parameters"] = {}
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        line_sv = line
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            nangles = int(line[0])
            break
        if len(line) > 2 and line[0].isdecimal() and line[1].isdecimal():
            # Looks like an off-diagonal!
            if len(line) != 8:
                raise ValueError(
                    f"Expected 6 element parameters, but found {len(line) - 2}"
                    f"\n\t{line_sv}"
                )
            key = (int(line[0]), int(line[1]))
            n_offdiagonals += 1
            odp = parameters["off-diagonal parameters"][key] = {}
            count = 0
            for val in line[2:]:
                odp[mdata[count]["parameter"]] = val
                count += 1
    if noffdiagonals != n_offdiagonals:
        logger.warning(
            f"Expected {noffdiagonals} off-diagonal parameters, "
            f"but found {n_offdiagonals}"
        )

    # Handle the angle parameters... the section looks like this:
    #
    # 62    ! Nr of angles;at1;at2;at3;Thetao,o;ka;kb;pv1;pv2;val(bo)
    #  2  2  2  74.0968  15.6349   3.6545   0.0000   0.2525   0.0000   1.2262
    #  3  3  3  62.0574  16.1545   3.0575   0.0000   0.0100   0.3556   1.0000
    #  2  3  3   2.5162  18.6626   4.8139   0.0000   0.0775   0.0000   1.0150
    # ...

    n_angles = 0
    parameters["angle parameters"] = {}
    mdata = metadata["angle parameters"]
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        line_sv = line
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            ntorsions = int(line[0])
            break
        if (
            len(line) > 3
            and line[0].isdecimal()
            and line[1].isdecimal()
            and line[2].isdecimal()
        ):
            # Looks like an angle!
            if len(line) != 10:
                raise ValueError(
                    f"Expected 10 angle parameters, but found {len(line) - 3}"
                    f"\n\t{line_sv}"
                )
            key = (int(line[0]), int(line[1]), int(line[2]))
            n_angles += 1
            ap = parameters["angle parameters"][key] = {}
            for count, val in enumerate(line[3:]):
                ap[mdata[count]["parameter"]] = val
    if nangles != n_angles:
        logger.warning(f"Expected {nangles} angle parameters, but found {n_angles}")

    # Handle the torsion parameters... the section looks like this:
    #
    # 14    ! Nr of torsions;at1;at2;at3;at4;;V1;V2;V3;V2(BO);vconj;n.u;n
    #  3  3  3  3   2.5000  32.8238   0.2056  -9.0000  -2.2982   0.0000   0.0000
    # ...
    # 0 element indices are wildcards.

    n_torsions = 0
    parameters["torsion parameters"] = {}
    mdata = metadata["torsion parameters"]
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        line_sv = line
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            nhbonds = int(line[0])
            break
        if (
            len(line) > 4
            and line[0].isdecimal()
            and line[1].isdecimal()
            and line[2].isdecimal()
            and line[3].isdecimal()
        ):
            # Looks like a torsion!
            if len(line) != 11:
                raise ValueError(
                    f"Expected 7 torsion parameters, but found {len(line) - 4}"
                    f"\n\t{line_sv}"
                )
            key = (int(line[0]), int(line[1]), int(line[2]), int(line[3]))
            n_torsions += 1
            tp = parameters["torsion parameters"][key] = {}
            for count, val in enumerate(line[4:]):
                tp[mdata[count]["parameter"]] = val
    if ntorsions != n_torsions:
        logger.warning(
            f"Expected {ntorsions} torsion parameters, but found {n_torsions}"
        )

    # Handle the hydrogen bond parameters... the section looks like this:
    #
    # 1    ! Nr of hydrogen bonds;at1;at2;at3;Rhb;Dehb;vhb1
    # 0  0  0   0.0000  -0.0000   0.0000  00.0000
    # ...

    n_hbonds = 0
    parameters["hydrogen bond parameters"] = {}
    mdata = metadata["hydrogen bond parameters"]
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        line_sv = line
        line = line.split("!")[0].split()
        if len(line) == 1 and line[0].isdecimal():
            logger.warning(f"Found another section after the hydrogen bonds: {line_sv}")
            break
        if (
            len(line) > 3
            and line[0].isdecimal()
            and line[1].isdecimal()
            and line[2].isdecimal()
        ):
            # Looks like a hydrogen bond!
            if len(line) != 7:
                raise ValueError(
                    f"Expected 4 hydrogen bond parameters, but found {len(line) - 3}"
                    f"\n\t{line_sv}"
                )
            # Zero indices used to indicate no hydrogen bond
            if line[0] == "0" and line[1] == "0" and line[2] == "0":
                nhbonds -= 1
                continue
            key = (int(line[0]), int(line[1]), int(line[2]))
            n_hbonds += 1
            hb = parameters["hydrogen bond parameters"][key] = {}
            for count, val in enumerate(line[3:]):
                hb[mdata[count]["parameter"]] = val
    if nhbonds != n_hbonds:
        logger.warning(
            f"Expected {nhbonds} hydrogen bond parameters, but found {n_hbonds}"
        )

    # See if there is anything else at the end of the file
    for line in it:
        if line.strip() == "" or line.strip()[0] == "!":
            continue
        logger.warning(f"Found another section after the hydrogen bonds: {line_sv}")

    return parameters


def read_reaxff_parameters(filename):
    """Read ReaxFF parameters from a file"""
    with open(filename, "r") as f:
        return parse_reax_data(f.readlines())


def reaxff_to_molssi_format(parameters, ffname="reaxff", version=None):
    """Create the text parameter file for ReaxFF

    Parameters
    ----------
    parameters : dict
        The forcefield parameters to write to the file.
    ffname : str = "reaxff"
        The name of the forcefield
    version : str = None
        The version number, defaults to todays date.

    Returns
    -------
    [str]
    """
    if version is None:
        version = date.isoformat(date.today()).replace("-", ".")

    #
    # The header and definition section
    #

    text = f"!MolSSI forcefield 1\n\n\n#define {ffname}\n\n"

    sections = []
    # The general section
    sections.append("reaxff_general_parameters")

    # The atomic parameter sections
    n_ap = len(metadata["atomic parameters"])
    for g0 in range(0, n_ap, 8):
        g1 = min(g0 + 8, n_ap)
        sections.append(f"reaxff_atomic_parameters_{g0 + 1}-{g1}")

    # The bond parameter sections
    n_bp = len(metadata["bond parameters"])
    for g0 in range(0, n_bp, 8):
        g1 = min(g0 + 8, n_ap)
        sections.append(f"reaxff_bond_parameters_{g0 + 1}-{g1}")

    # The off-diagonal parameter section
    sections.append("reaxff_off-diagonal_parameters")

    # The angle parameter section
    sections.append("reaxff_angle_parameters")

    # The torsion parameter section
    sections.append("reaxff_torsion_parameters")

    # The hydrogen bond parameter sections
    if len(parameters["hydrogen bond parameters"]) > 0:
        sections.append("reaxff_hydrogen-bond_parameters")

    # And tabulate
    table = {
        "Version": [version] * len(sections),
        "Ref": ["1"] * len(sections),
        "Section": sections,
        "Label": [ffname] * len(sections),
    }

    tmp = tabulate(
        table,
        headers="keys",
        tablefmt="simple",
        colalign=("center", "right", "left", "left"),
        disable_numparse=False,
    ).splitlines()
    tmp[0] = "!" + tmp[0][1:]
    tmp[1] = "!" + tmp[1][1:]
    text += "\n".join(tmp)

    #
    # The general parameters
    #

    text += f"\n\n\n#reaxff_general_parameters {ffname}\n\n"

    tmp = [(v["parameter"], v["description"]) for v in metadata["general parameters"]]
    tmp = sorted(tmp, key=lambda k: k[0])

    gp = parameters["general parameters"]
    table = {
        "Version": [version] * len(tmp),
        "Ref": ["1"] * len(tmp),
        "Parameter": [v[0] for v in tmp],
        "Value": [gp[v[0]] for v in tmp],
        "Description": [v[1] for v in tmp],
    }

    tmp = tabulate(
        table,
        headers="keys",
        tablefmt="simple",
        colalign=("center", "right", "center", "decimal", "left"),
        disable_numparse=False,
    ).splitlines()
    tmp[0] = "!" + tmp[0][1:]
    tmp[1] = "!" + tmp[1][1:]
    text += "\n".join(tmp)

    #
    # The atomic parameters, split into groups of 8
    #

    tmp = sorted([v["parameter"] for v in metadata["atomic parameters"]])

    # Get the elements and order by atomic number
    ap = parameters["atomic parameters"]
    atomic_metadata = [
        (element_data[s]["atomic number"], s, index)
        for index, s in enumerate(ap.keys(), start=1)
    ]
    atomic_metadata = sorted(atomic_metadata, key=lambda k: k[0])

    # Create a mapping from old index to new for the bond terms, etc.
    to_new = {data[2]: index for index, data in enumerate(atomic_metadata)}

    # Ordered list of symbols
    symbols = [s[1] for s in atomic_metadata]

    # Groups of 8
    n_a = len(symbols)
    for g0 in range(0, n_ap, 8):
        g1 = min(g0 + 8, n_ap)
        text += f"\n\n\n#reaxff_atomic_parameters_{g0 + 1}-{g1} {ffname}\n\n"
        table = {
            "Version": [version] * n_a,
            "Ref": ["1"] * n_a,
            "Center": symbols,
        }
        for parameter in tmp[g0:g1]:
            table[parameter] = [ap[s][parameter] for s in symbols]
        lines = tabulate(
            table,
            headers="keys",
            tablefmt="simple",
            colalign=["center", "right", "center"] + ["decimal"] * (g1 - g0),
            disable_numparse=False,
        ).splitlines()
        lines[0] = "!" + lines[0][1:]
        lines[1] = "!" + lines[1][1:]
        text += "\n".join(lines)

    #
    # The bond parameters, split into groups of 8
    #

    tmp = sorted([v["parameter"] for v in metadata["bond parameters"]])

    # Get the elements and order by atomic number
    bp = parameters["bond parameters"]

    # The indices, and turn into symbols
    indices = []
    for i, j in bp.keys():
        i1 = to_new[i]
        j1 = to_new[j]
        if i1 < j1:
            i1, j1 = j1, i1
        indices.append((i1, j1, (i, j)))
    indices = sorted(indices, key=lambda k: k[1])
    indices = sorted(indices, key=lambda k: k[0])

    # Groups of 8
    n = len(bp)
    for g0 in range(0, n_bp, 8):
        g1 = min(g0 + 8, n_bp)
        text += f"\n\n\n#reaxff_bond_parameters_{g0 + 1}-{g1} {ffname}\n\n"
        table = {
            "Version": [version] * n,
            "Ref": ["1"] * n,
            "I": [symbols[k[0]] for k in indices],
            "J": [symbols[k[1]] for k in indices],
        }
        for parameter in tmp[g0:g1]:
            table[parameter] = [bp[k[2]][parameter] for k in indices]
        lines = tabulate(
            table,
            headers="keys",
            tablefmt="simple",
            colalign=["center", "right", "center", "center"] + ["decimal"] * (g1 - g0),
            disable_numparse=False,
        ).splitlines()
        lines[0] = "!" + lines[0][1:]
        lines[1] = "!" + lines[1][1:]
        text += "\n".join(lines)

    #
    # The off-diagonal parameters
    #

    tmp = sorted([v["parameter"] for v in metadata["off-diagonal parameters"]])

    # Get the elements and order by atomic number
    odp = parameters["off-diagonal parameters"]
    if len(odp) > 0:
        # The indices, and turn into symbols
        indices = []
        for i, j in odp.keys():
            i1 = to_new[i]
            j1 = to_new[j]
            if i1 < j1:
                i1, j1 = j1, i1
            indices.append((i1, j1, (i, j)))
        indices = sorted(indices, key=lambda k: k[1])
        indices = sorted(indices, key=lambda k: k[0])

        n = len(odp)
        text += f"\n\n\n#reaxff_off-diagonal_parameters {ffname}\n\n"
        table = {
            "Version": [version] * n,
            "Ref": ["1"] * n,
            "I": [symbols[k[0]] for k in indices],
            "J": [symbols[k[1]] for k in indices],
        }
        for parameter in tmp:
            table[parameter] = [odp[k[2]][parameter] for k in indices]
        lines = tabulate(
            table,
            headers="keys",
            tablefmt="simple",
            colalign=["center", "right", "center", "center"] + ["decimal"] * len(tmp),
            disable_numparse=False,
        ).splitlines()
        lines[0] = "!" + lines[0][1:]
        lines[1] = "!" + lines[1][1:]
        text += "\n".join(lines)

    #
    # The angle parameters
    #

    tmp = sorted([v["parameter"] for v in metadata["angle parameters"]])

    # Get the elements and order by atomic number
    ap = parameters["angle parameters"]

    # The indices, and turn into symbols
    indices = []
    for i, j, k in ap.keys():
        i1 = to_new[i]
        j1 = to_new[j]
        k1 = to_new[k]
        if i1 < k1:
            i1, k1 = k1, i1
        indices.append((i1, j1, k1, (i, j, k)))
    indices = sorted(indices, key=lambda k: k[2])
    indices = sorted(indices, key=lambda k: k[0])
    indices = sorted(indices, key=lambda k: k[1])

    n = len(ap)
    text += f"\n\n\n#reaxff_angle_parameters {ffname}\n\n"
    table = {
        "Version": [version] * n,
        "Ref": ["1"] * n,
        "I": [symbols[k[0]] for k in indices],
        "J": [symbols[k[1]] for k in indices],
        "K": [symbols[k[2]] for k in indices],
    }
    for parameter in tmp:
        table[parameter] = [ap[k[3]][parameter] for k in indices]
    lines = tabulate(
        table,
        headers="keys",
        tablefmt="simple",
        colalign=["center", "right", "center", "center"] + ["decimal"] * len(tmp),
        disable_numparse=False,
    ).splitlines()
    lines[0] = "!" + lines[0][1:]
    lines[1] = "!" + lines[1][1:]
    text += "\n".join(lines)

    #
    # The torsion parameters
    #

    tmp = sorted([v["parameter"] for v in metadata["torsion parameters"]])

    # Get the elements and order by atomic number
    tp = parameters["torsion parameters"]

    # The indices, and turn into symbols
    indices = []
    for i, j, k, l in tp.keys():
        i1 = 999 if i == 0 else to_new[i]
        j1 = to_new[j]
        k1 = to_new[k]
        l1 = 999 if i == 0 else to_new[l]
        if j1 < k1 or j1 == k1 and i1 < l1:
            i1, j1, k1, l1 = l1, k1, j1, i1
        indices.append((i1, j1, k1, l1, (i, j, k, l)))
    indices = sorted(indices, key=lambda k: k[3])
    indices = sorted(indices, key=lambda k: k[0])
    indices = sorted(indices, key=lambda k: k[2])
    indices = sorted(indices, key=lambda k: k[1])

    n = len(tp)
    text += f"\n\n\n#reaxff_torsion_parameters {ffname}\n\n"
    table = {
        "Version": [version] * n,
        "Ref": ["1"] * n,
        "I": ["*" if k[0] == 999 else symbols[k[0]] for k in indices],
        "J": [symbols[k[1]] for k in indices],
        "K": [symbols[k[2]] for k in indices],
        "L": ["*" if k[3] == 999 else symbols[k[3]] for k in indices],
    }
    for parameter in tmp:
        table[parameter] = [tp[k[4]][parameter] for k in indices]
    lines = tabulate(
        table,
        headers="keys",
        tablefmt="simple",
        colalign=["center", "right", "center", "center"] + ["decimal"] * len(tmp),
        disable_numparse=False,
    ).splitlines()
    lines[0] = "!" + lines[0][1:]
    lines[1] = "!" + lines[1][1:]
    text += "\n".join(lines)

    #
    # The hydrogen-bond parameters. They are directional, no symmetry.
    #

    tmp = sorted([v["parameter"] for v in metadata["hydrogen bond parameters"]])

    # Get the elements and order by atomic number
    hbp = parameters["hydrogen bond parameters"]

    if len(hbp) > 0:
        # The indices, and turn into symbols
        indices = []
        for i, j, k in hbp.keys():
            i1 = to_new[i]
            j1 = to_new[j]
            k1 = to_new[k]
            indices.append((i1, j1, k1, (i, j, k)))
        indices = sorted(indices, key=lambda k: k[2])
        indices = sorted(indices, key=lambda k: k[0])
        indices = sorted(indices, key=lambda k: k[1])

        n = len(hbp)
        text += f"\n\n\n#reaxff_hydrogen-bond_parameters {ffname}\n\n"
        table = {
            "Version": [version] * n,
            "Ref": ["1"] * n,
            "I": [symbols[k[0]] for k in indices],
            "J": [symbols[k[1]] for k in indices],
            "K": [symbols[k[2]] for k in indices],
        }
        for parameter in tmp:
            table[parameter] = [hbp[k[3]][parameter] for k in indices]
        lines = tabulate(
            table,
            headers="keys",
            tablefmt="simple",
            colalign=["center", "right", "center", "center"] + ["decimal"] * len(tmp),
            disable_numparse=False,
        ).splitlines()
        lines[0] = "!" + lines[0][1:]
        lines[1] = "!" + lines[1][1:]
        text += "\n".join(lines)

    text += f"""

#reference 1
@Author SEAMM
@Date {date.isoformat(date.today())}
ReaxFF forcefield translated from Reax format

{parameters["comment"]}

#end
"""

    return text


def write_ff_file(path, parameters, ffname=None, version=None):
    """Write the MolSSI forcefield to the given path

    Parameters
    ----------
    path : str or pathlib.Path
        The path to write the forcefield to
    parameters : dict
        The forcefield parameters
    ffname : str = None
        The name of the forcefield. The default is to use the stem of the path
    version : str = None
        The version for the forcefield. The default is today's date
    """
    path = Path(path)
    if ffname is None:
        ffname = path.stem

    text = reaxff_to_molssi_format(parameters, ffname, version)
    path.write_text(text)


def run():
    """Run the script from the commandline."""
    # Create the argument parser and set the debug level ASAP
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--log-level",
        default="WARNING",
        type=str.upper,
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="The level of informational output, defaults to '%(default)s'",
    )

    # set up the subparsers
    subparser = parser.add_subparsers()

    # The import subparser
    import_parser = subparser.add_parser("import", help="Import a ReaxFF forcefield")

    import_parser.set_defaults(func=import_reaxff)

    import_parser.add_argument(
        "reaxff",
        default=None,
        help="The ReaxFF file to read.",
    )
    import_parser.add_argument(
        "output",
        default=None,
        help="The forcefield file to write.",
    )
    import_parser.add_argument(
        "--personal",
        action="store_true",
        help="Use the personal forcefield database.",
    )
    parser.add_argument(
        "--version",
        default=None,
        help="Forcefield version, defaults to todays date",
    )
    parser.add_argument(
        "--ffname",
        default=None,
        help="Name of the forcefield, defaults to the stem of the file name",
    )

    # Parse the command-line arguments and call the requested function or the GUI
    options = parser.parse_args()

    # Set up the logging
    logging.basicConfig(level=options.log_level)

    # do it!
    if "func" not in options:
        print(f"Missing arguments to the ReaxFF handler: {' '.join(sys.argv[1:])}")
        # Append help so help will be printed
        sys.argv.append("--help")
        # re-run
        run()
    else:
        sys.exit(options.func(options))


class ReaxFFMixin:

    def eex_reaxff_general_parameters(self, eex, configuration):
        """Create the custom forcefield file for the ReaxFF calculation."""
        n_atoms = configuration.n_atoms

        # Need eex["charges"] to write out the correct structure for LAMMPS
        eex["charges"] = [0.0] * n_atoms

        # pprint.pprint(eex)
        # print("\n" * 5)
        # pprint.pprint(self.ff)

        # elements in correct order
        symbols = eex["atom types"]

        # And the forcefield file
        lines = []
        lines.append(
            f"ReaxFF forcefield {self.current_forcefield} for "
            f"{configuration.system.name}/{configuration.name}"
        )

        # General parameters
        P = self.ff["reaxff_general_parameters"]
        lines.append(" 39        ! Number of general parameters")
        for mdata in metadata["general parameters"]:
            parameter = mdata["parameter"]
            D = P[parameter]
            lines.append(
                f" {float(D['value']):9.4f} ! {parameter}: {mdata['description']}"
            )

        # Atomic parameters
        M = metadata["atomic parameters"]
        n = len(M)
        n_types = len(symbols)
        line = f" {n_types:2d}        ! Num atom params; "
        tmp = [d["parameter"] for d in M[0:8]]
        line += "; ".join(tmp)
        lines.append(line)
        for g0 in range(8, n, 8):
            g1 = min(g0 + 8, n)
            tmp = [d["parameter"] for d in M[g0:g1]]
            lines.append("           ! " + "; ".join(tmp))

        # Eat for the atomic energy correction
        Eat = {}
        for S in symbols:
            P = {}
            for section in eex["terms"]["reaxff atomic parameters"]:
                P.update(self.ff[section][(S,)])

            Eat[S] = P["Hat"]
            for g0 in range(0, n, 8):
                g1 = min(g0 + 8, n)
                tmp = [f"{float(P[k['parameter']]):8.4f}" for k in M[g0:g1]]
                if g0 == 0:
                    lines.append(f" {S:2} " + " ".join(tmp))
                else:
                    lines.append("    " + " ".join(tmp))

        # Calculate the atom offset energy
        E = sum([float(Eat[S]) for S in configuration.atoms.symbols])
        eex["Sum of atomic energies"] = E

        # Bond parameters
        M = metadata["bond parameters"]
        n = len(M)

        # How many bonds will we have?
        section = eex["terms"]["reaxff bond parameters"][0]
        indices = []
        for key in self.ff[section]:
            i, j = key
            if i in symbols and j in symbols:
                indices.append(key)

        line = f"{len(indices):3d}        ! Num bonds; "
        tmp = [d["parameter"] for d in M[0:8]]
        line += "; ".join(tmp)
        lines.append(line)
        for g0 in range(8, n, 8):
            g1 = min(g0 + 8, n)
            tmp = [d["parameter"] for d in M[g0:g1]]
            lines.append("           ! " + "; ".join(tmp))

        for key in indices:
            P = {}
            for section in eex["terms"]["reaxff bond parameters"]:
                P.update(self.ff[section][key])

            i, j = key
            i = symbols.index(i) + 1
            j = symbols.index(j) + 1
            for g0 in range(0, n, 8):
                g1 = min(g0 + 8, n)
                tmp = [f"{float(P[k['parameter']]):8.4f}" for k in M[g0:g1]]
                if g0 == 0:
                    lines.append(f"{int(i):3d}{int(j):3d} " + " ".join(tmp))
                else:
                    lines.append("       " + " ".join(tmp))

        # Off-Diagonal parameters
        M = metadata["off-diagonal parameters"]
        n = len(M)

        # How many off-diagonals will we have?
        section = eex["terms"]["reaxff off-diagonal parameters"][0]
        indices = []
        for key in self.ff[section]:
            i, j = key
            if i in symbols and j in symbols:
                indices.append(key)

        line = f"{len(indices):3d}        ! Num off-diagonals; "
        tmp = [d["parameter"] for d in M[0:8]]
        line += "; ".join(tmp)
        lines.append(line)
        for g0 in range(8, n, 8):
            g1 = min(g0 + 8, n)
            tmp = [d["parameter"] for d in M[g0:g1]]
            lines.append("           ! " + "; ".join(tmp))

        for key in indices:
            P = {}
            for section in eex["terms"]["reaxff off-diagonal parameters"]:
                P.update(self.ff[section][key])

            i, j = key
            i = symbols.index(i) + 1
            j = symbols.index(j) + 1
            for g0 in range(0, n, 8):
                g1 = min(g0 + 8, n)
                tmp = [f"{float(P[k['parameter']]):8.4f}" for k in M[g0:g1]]
                if g0 == 0:
                    lines.append(f"{int(i):3d}{int(j):3d} " + " ".join(tmp))
                else:
                    lines.append("       " + " ".join(tmp))

        # Angle parameters
        M = metadata["angle parameters"]
        n = len(M)

        # How many angles will we have?
        section = eex["terms"]["reaxff angle parameters"][0]
        indices = []
        for key in self.ff[section]:
            i, j, k = key
            if i in symbols and j in symbols and k in symbols:
                indices.append(key)

        line = f"{len(indices):3d}        ! Num angles; "
        tmp = [d["parameter"] for d in M[0:8]]
        line += "; ".join(tmp)
        lines.append(line)
        for g0 in range(8, n, 8):
            g1 = min(g0 + 8, n)
            tmp = [d["parameter"] for d in M[g0:g1]]
            lines.append("           ! " + "; ".join(tmp))

        for key in indices:
            P = {}
            for section in eex["terms"]["reaxff angle parameters"]:
                P.update(self.ff[section][key])

            i, j, k = key
            i = symbols.index(i) + 1
            j = symbols.index(j) + 1
            k = symbols.index(k) + 1
            for g0 in range(0, n, 8):
                g1 = min(g0 + 8, n)
                tmp = [f"{float(P[k['parameter']]):8.4f}" for k in M[g0:g1]]
                if g0 == 0:
                    lines.append(f"{int(i):3d}{int(j):3d}{int(k):3d} " + " ".join(tmp))
                else:
                    lines.append("          " + " ".join(tmp))

        # Torsion parameters
        M = metadata["torsion parameters"]
        n = len(M)

        # How many torsions will we have?
        section = eex["terms"]["reaxff torsion parameters"][0]
        indices = []
        for key in self.ff[section]:
            i, j, k, l = key  # noqa:E741
            if i in symbols and j in symbols and k in symbols and l in symbols:
                indices.append(key)

        line = f"{len(indices):3d}        ! Num torsions; "
        tmp = [d["parameter"] for d in M[0:8]]
        line += "; ".join(tmp)
        lines.append(line)
        for g0 in range(8, n, 8):
            g1 = min(g0 + 8, n)
            tmp = [d["parameter"] for d in M[g0:g1]]
            lines.append("           ! " + "; ".join(tmp))

        for key in indices:
            P = {}
            for section in eex["terms"]["reaxff torsion parameters"]:
                P.update(self.ff[section][key])

            i, j, k, l = key  # noqa:E741
            i = 0 if i == "*" else symbols.index(i) + 1
            j = symbols.index(j) + 1
            k = symbols.index(k) + 1
            l = 0 if l == "*" else symbols.index(l) + 1  # noqa:E741
            for g0 in range(0, n, 8):
                g1 = min(g0 + 8, n)
                tmp = [f"{float(P[k['parameter']]):8.4f}" for k in M[g0:g1]]
                if g0 == 0:
                    lines.append(
                        f"{int(i):3d}{int(j):3d}{int(k):3d}{int(l):3d} " + " ".join(tmp)
                    )
                else:
                    lines.append("             " + " ".join(tmp))

        # Hydrogen Bond parameters
        M = metadata["hydrogen bond parameters"]
        n = len(M)

        # How many hydrogen bonds will we have?
        if "reaxff hydrogen bond parameters" not in eex["terms"]:
            indices = []
        else:
            section = eex["terms"]["reaxff hydrogen bond parameters"][0]
            indices = []
            for key in self.ff[section]:
                i, j, k = key
                if i in symbols and j in symbols and k in symbols:
                    indices.append(key)

        line = f"{len(indices):3d}        ! Num hydrogen bonds; "
        tmp = [d["parameter"] for d in M[0:8]]
        line += "; ".join(tmp)
        lines.append(line)
        for g0 in range(8, n, 8):
            g1 = min(g0 + 8, n)
            tmp = [d["parameter"] for d in M[g0:g1]]
            lines.append("           ! " + "; ".join(tmp))

        if "reaxff hydrogen bond parameters" not in eex["terms"]:
            for key in indices:
                P = {}
                for section in eex["terms"]["reaxff hydrogen bond parameters"]:
                    P.update(self.ff[section][key])

                i, j, k = key
                i = symbols.index(i) + 1
                j = symbols.index(j) + 1
                k = symbols.index(k) + 1
                for g0 in range(0, n, 8):
                    g1 = min(g0 + 8, n)
                    tmp = [f"{float(P[k['parameter']]):8.4f}" for k in M[g0:g1]]
                    if g0 == 0:
                        lines.append(
                            f"{int(i):3d}{int(j):3d}{int(k):3d} " + " ".join(tmp)
                        )
                    else:
                        lines.append("          " + " ".join(tmp))

        eex["forcefield"] = "\n".join(lines)

    def eex_reaxff_atomic_parameters(self, eex, configuration):
        """Not used: everything is handled in the general parameters code."""
        pass

    def eex_reaxff_bond_parameters(self, eex, configuration):
        """Not used: everything is handled in the general parameters code."""
        pass

    def eex_reaxff_off_diagonal_parameters(self, eex, configuration):
        """Not used: everything is handled in the general parameters code."""
        pass

    def eex_reaxff_angle_parameters(self, eex, configuration):
        """Not used: everything is handled in the general parameters code."""
        pass

    def eex_reaxff_torsion_parameters(self, eex, configuration):
        """Not used: everything is handled in the general parameters code."""
        pass

    def eex_reaxff_hydrogen_bond_parameters(self, eex, configuration):
        """Not used: everything is handled in the general parameters code."""
        pass

    def _parse_biosym_reaxff_general_parameters(self, data):
        """
        Process general parameters for ReaxFF

        #reaxff_general_parameters ci-reaxff_CH_2018

        !Version      Ref   Parameter       Value  Description
        !---------  -----  ------------  --------  -------------------------------------
        2025.03.21      1   BO_cutoff      0.1     Cutoff for bond order (*100)
        2025.03.21      1     Pboc,1      50       Overcoordination parameter 1
        2025.03.21      1     Pboc,2       9.5469  Overcoordination parameter 2
        2025.03.21      1     Pcoa,2      26.5405  Valency angle conjugation parameter
        ...
        """  # nopep8
        section = data["section"]
        label = data["label"]

        if section not in self.data:
            self.data[section] = {}
        if label in self.data[section]:
            msg = "'{}' already defined in section '{}'".format(label, section)
            logger.error(msg)
            raise RuntimeError(msg)
        self.data[section][label] = data
        gp = self.data[section][label]["parameters"] = {}

        for line in data["lines"]:
            version, reference, parameter, value, description = line.split(maxsplit=4)

            if parameter not in gp:
                gp[parameter] = {}

            V = packaging.version.Version(version)
            if V in gp[parameter]:
                msg = (
                    f"Reax general parameter '{parameter}' defined more than "
                    f"once in section '{section}'!\n\n{line}"
                )
                logger.error(msg)
                raise RuntimeError(msg)
            gp[parameter][V] = {
                "value": value,
                "description": description,
            }

        if not self.keep_lines:
            del data["lines"]


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-

"""Mixin class for handling Dreiding forcefields"""

import logging
from math import sqrt, sin, radians  # noqa: F401
import pprint  # noqa: F401

logger = logging.getLogger(__name__)
# logger.setLevel("DEBUG")


class DreidingMixin:

    def eex_dreiding_atomic_parameters(self, eex, configuration):
        """Get the charges for the structure

        If they do not exists on the structure, they are created
        using the bond increments and saved on the structure"""

        logger.debug("entering eex_dreiding_atomic_parameters")

        # Get the bonds, angles, ...
        self.dreiding_bonds(eex, configuration)
        self.dreiding_angles(eex, configuration)
        self.dreiding_dihedrals(eex, configuration)

    def dreiding_bonds(self, eex, configuration):
        """Get the bonds and parameters for the structure."""

        logger.debug("entering dreiding_bonds")

        types = self.topology["types"]
        bonds = self.topology["bonds"]
        bond_orders = self.topology["bond orders"]
        atomic_parameters = self.ff["dreiding_atomic_parameters"]

        # pprint.pp(atomic_parameters)

        # Add the bonds to the eex
        terms = eex["terms"]
        if "bonds" not in terms:
            terms["bonds"] = []
        if "quadratic_bond" not in terms["bonds"]:
            terms["bonds"].append("quadratic_bond")

        if "bonds" not in eex:
            eex["bonds"] = []
            eex["bond parameters"] = []

        result = eex["bonds"]
        parameters = eex["bond parameters"]

        for i, j in bonds:
            itype = types[i]
            jtype = types[j]
            key, flipped = self.make_canonical("like_bond", (itype, jtype))

            parameters_type = "explicit"
            form = "quadratic_bond"
            real_types = (itype, jtype)

            n = bond_orders[(i, j)]
            if n == 5:
                # Aromatic
                n = 1.5

            R_i = float(atomic_parameters[(itype,)]["Radius"])
            R_j = float(atomic_parameters[(jtype,)]["Radius"])
            R0 = R_i + R_j - 0.01
            K2 = n * 700
            # D = n * 70
            # alpha = sqrt(K2/(2*D))

            parameter_values = {"K2": f"{K2 / 2:.4f}", "R0": f"{R0:.4f}"}

            new_value = (
                form,
                parameter_values,
                (types[i], types[j]),
                parameters_type,
                real_types,
            )

            try:
                index = parameters.index(new_value) + 1
            except ValueError:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, index))
        eex["n_bonds"] = len(result)
        eex["n_bond_types"] = len(parameters)

    def dreiding_angles(self, eex, configuration):
        """Get the angles and parameters for the structure."""

        logger.debug("entering dreiding_angles")

        types = self.topology["types"]
        angles = self.topology["angles"]
        atomic_parameters = self.ff["dreiding_atomic_parameters"]

        # Add the angless to the eex
        terms = eex["terms"]
        if "angle" not in terms:
            terms["angle"] = []
        if "cosine/squared" not in terms["angle"]:
            terms["angle"].append("cosine/squared")
        if "cosine" not in terms["angle"]:
            terms["angle"].append("cosine")

        if "angles" not in eex:
            eex["angles"] = []
            eex["angle parameters"] = []

        result = eex["angles"]
        parameters = eex["angle parameters"]

        for i, j, k in angles:
            itype = types[i]
            jtype = types[j]
            ktype = types[k]
            key, flipped = self.make_canonical("like_angle", (itype, jtype, ktype))

            parameters_type = "explicit"
            real_types = (itype, jtype, ktype)

            theta0 = float(atomic_parameters[(jtype,)]["Theta0"])

            if theta0 == 180:
                form = "cosine"
                K2 = 100.0
                parameter_values = {"K2": f"{K2 / 2:.4f}"}
            else:
                form = "cosine/squared"
                K2 = 100.0 / sin(radians(theta0)) ** 2
                parameter_values = {"K2": f"{K2 / 2:.4f}", "Theta0": f"{theta0:.4f}"}

            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k]),
                parameters_type,
                real_types,
            )

            try:
                index = parameters.index(new_value) + 1
            except ValueError:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, index))
        eex["n_angles"] = len(result)
        eex["n_angle_types"] = len(parameters)

    def dreiding_dihedrals(self, eex, configuration):
        """Get the dihedrals and parameters for the structure."""

        logger.debug("entering dreiding_dihedrals")

        types = self.topology["types"]
        torsions = self.topology["torsions"]
        bond_orders = self.topology["bond orders"]

        # pprint.pp(atomic_parameters)

        # Add the dihedrals to the eex
        terms = eex["terms"]
        if "torsion" not in terms:
            terms["torsion"] = []
        if "torsion_charmm" not in terms["torsion"]:
            terms["torsion"].append("torsion_charmm")

        if "torsions" not in eex:
            eex["torsions"] = []
            eex["torsion parameters"] = []

        result = eex["torsions"]
        parameters = eex["torsion parameters"]

        for i, j, k, l in torsions:
            itype = types[i]
            jtype = types[j]
            ktype = types[k]
            ltype = types[l]
            key, flipped = self.make_canonical(
                "like_torsion", (itype, jtype, ktype, ltype)
            )

            ihyb = itype[2] if len(itype) > 2 else 0
            jhyb = jtype[2] if len(jtype) > 2 else 0
            khyb = ktype[2] if len(ktype) > 2 else 0
            lhyb = ltype[2] if len(ltype) > 2 else 0
            ihyb = 2 if ihyb == "R" else 0 if ihyb == "_" else int(ihyb)
            jhyb = 2 if jhyb == "R" else 0 if jhyb == "_" else int(jhyb)
            khyb = 2 if khyb == "R" else 0 if khyb == "_" else int(khyb)
            lhyb = 2 if lhyb == "R" else 0 if lhyb == "_" else int(lhyb)

            bond_order = bond_orders[(j, k)]
            if bond_order == 5:
                # Aromatic
                bond_order = 1.5

            parameters_type = "explicit"
            form = "torsion_charmm"
            real_types = (itype, jtype, ktype, ltype)

            # An expert system, as written in the paper:
            # (a) A dihedral single bond involving two sp3 atoms (J,K = X_3)
            if jhyb == 3 and khyb == 3:
                V = 2.0
                n = 3
                phi0 = 180

                # (h) A dihedral single bond involving two sp3 atoms of the oxygen
                #     column (J,K = X.3 of column 16)
                if jtype[0:2] in ("O_", "S_", "Se", "Te") and ktype[0:2] in (
                    "O_",
                    "S_",
                    "Se",
                    "Te",
                ):
                    V = 1.0
                    n = 6
                    phi0 = 90

            # (b) A dihedral single bond involving one sp2 center and one sp3center
            #     e.g., the C-C bond in acetic acid [CH3C(0)^0H)]
            #     (J= X.2, X.R; K = X.3)
            elif jhyb == 3 and khyb == 2 and bond_order == 1:
                V = 1.0
                n = 6
                phi0 = 0.0

                # (i) For dihedral bonds involving an sp3 atom of the oxygen column with
                #     an sp2 or resonant atom of another column, the pair and the
                #     oxygen-like prefers to overlap the orbitals of the sp2 atom,
                #     leading to a planar configuration
                #     (J = X.3 of column16, K = X.2,X.R)
                if (jtype[0:2] in ("O_", "S_", "Se", "Te") and khyb == 2) or (
                    ktype[0:2] in ("O_", "S_", "Se", "Te") and jhyb == 2
                ):
                    V = 2.0
                    n = 2
                    phi0 = 180

                # (j) An exception to the above principles is made for the case of a
                #  dihedral single bond involving one sp2 atom (J = X.2, X.R) and one
                #  sp3 atom (K = X.3) (for example, the single bond of propene). The
                #   problem here is that for a system such as propene there is a 3-fold
                #   barrier with the sp3 center eclipsing the double bond, whereas for
                #   the CC bond, say, of acetate anion (CH3-C-00"), the barrier should
                #   have 6-fold character. To accommodate this we use (15) unless I is
                #   not an sp2 center (X.2 or X.R); otherwise we use the following
                #   I ^ X.2, X.R; J = X.2,X.R; K = X.3
                if (jhyb == 2 and ihyb != 2) or (khyb == 2 and lhyb != 2):
                    V = 2.0
                    n = 3
                    phi0 = 180

            # (c) A dihedral double bond involving two sp2 atoms (J,K = X.2)
            elif jhyb == 2 and khyb == 2 and bond_order == 2:
                V = 45.0
                n = 2
                phi0 = 180

            # (d) A dihedral resonance bond (bond order = 1.5) involving
            #     two resonant atoms (J,K = X.R)
            # NB. This is an issue for aromatic rings with single and double bonds.
            elif jhyb == 2 and khyb == 2 and bond_order == 1.5:
                V = 25.0
                n = 2
                phi0 = 180

            # (e) A dihedral single bond involving two sp2 or resonant atoms
            #     [e.g., the middle bond of butadiene] {J,K = X.2, X.R)
            elif jhyb == 2 and khyb == 2 and bond_order == 1:
                V = 5
                n = 2
                phi0 = 180

                # (f) An exception to (e) is that for an exocyclic dihedral single bond
                #     involving two aromatic atoms [e.g., a phenyl ester or the bond
                #     between the two rings of biphenyl] (J,K = X.R)
                # NB. This is an issue for aromatic rings with single and double bonds.
                if (len(jtype) > 2 and jtype[2] == "R") and (
                    len(ktype) > 2 and ktype[2] == "R"
                ):
                    V = 10.0
            else:
                # linear bond, like acetylene, so no torsion
                continue

            # Normalize by the number of dihedrals sharing the central bond
            n_bonds = self.topology["bonds_from_atom"]
            n_j = len(n_bonds[j])
            n_k = len(n_bonds[k])
            V /= n_j * n_k

            # Need to multiply by n and phase of 180 for LAMMPS/charmm
            phi0 = (n * phi0 + 180) % 360
            parameter_values = {
                "n": n,
                "Phi0": int(phi0),
                "K": f"{V / 2:.4f}",
                "weight": 0.0,
            }

            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )

            try:
                index = parameters.index(new_value) + 1
            except ValueError:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_torsions"] = len(result)
        eex["n_torsion_types"] = len(parameters)

    def dreiding_hydrogen_bonds(self, eex, configuration):
        """Rework the nonbonds and add the hydrogen bonds, if needed."""

        logger.debug("entering dreiding_bonds")

        types = self.topology["types"]

        # If there are no possible hydrogen bonds, leave the nonbonds as is
        if "H__HB" not in types:
            return

        # Otherwise we need to create the full triangle of non bonds and add the
        # hydrogen bond potentials for pairs that can have hydrogen bonds

        # What atom types are the H__HB bonded to?
        atom_types = self.topology["types"]
        bonds = self.topology["bonds"]
        donors = set()
        for i, j in bonds:
            itype = atom_types[i]
            jtype = atom_types[j]
            if itype == "H__HB":
                donors.add(jtype)
            elif jtype == "H__HB":
                donors.add(itype)
        acceptors = set()
        for itype in ("N_3", "O_3", "F_"):
            if itype in atom_types:
                acceptors.add(itype)

        eex["nonbonds"] = []
        parameters = eex["nonbond parameters"] = []

        types = eex["atom types"]
        h_index = types.index("H__HB") + 1
        for i, itype in enumerate(types, start=1):
            (
                i_parameters_type,
                i_real_types,
                i_form,
                i_parameter_values,
            ) = self.nonbond_parameters(itype, form="nonbond(12-6)")
            i_rmin = i_parameter_values["sigma"]
            i_eps = i_parameter_values["eps"]
            for j, jtype in enumerate(types[0:i], start=1):
                (
                    j_parameters_type,
                    j_real_types,
                    j_form,
                    j_parameter_values,
                ) = self.nonbond_parameters(itype, form="nonbond(12-6)")
                j_rmin = j_parameter_values["sigma"]
                j_eps = j_parameter_values["eps"]
                parameter_values = {
                    "sigma": sqrt(i_rmin * j_rmin),
                    "eps": sqrt(i_eps * j_eps),
                }
                new_value = (
                    i_form,
                    parameter_values,
                    (j, i),
                    (jtype, itype),
                    (j_real_types[0], i_real_types[0]),
                )
                parameters.append(new_value)

                if itype in donors or jtype in donors:
                    # Is there a possible hydrogen bond?
                    if jtype in donors and itype in acceptors:
                        new_value = (
                            "hbond/dreiding/lj",
                            {
                                "h_index": h_index,
                                "donor flag": "i",
                                "eps": 9.0,
                                "sigma": 2.75,
                                "exponent": 4,
                            },
                            (j, i),
                            (jtype, itype),
                            (jtype, itype),
                        )
                    elif jtype in acceptors and itype in donors:
                        new_value = (
                            "hbond/dreiding/lj",
                            {
                                "h_index": h_index,
                                "donor flag": "j",
                                "eps": 9.0,
                                "sigma": 2.75,
                                "exponent": 4,
                            },
                            (j, i),
                            (jtype, itype),
                            (jtype, itype),
                        )
                    parameters.append(new_value)



# -*- coding: utf-8 -*-

import logging
import pprint  # noqa: F401

logger = logging.getLogger(__name__)


class EEX_Mixin:
    """A mixin class for handling the energy expression (eex)."""

    def eex_atomic_charge(self, eex, configuration):
        """Handle charges."""
        self.eex_bond_charge_increment(eex, configuration)

    def eex_charge(self, eex, configuration):
        """Do nothing routine since charges are handled by the increments."""
        pass

    def eex_bond_charge_increment(self, eex, configuration):
        """Get the charges for the structure

        If they do not exist on the structure, they are created
        using the bond increments and saved on the structure"""

        logger.debug("entering eex_increment")

        # terms = self.terms
        ff_name = self.current_forcefield
        atoms = configuration.atoms
        # if "atomic charge" in terms and "shell model" in terms:
        #     logger.debug("Getting the charges for the system")

        #     # Atom types
        #     key = f"atom_types_{ff_name}"
        #     atom_types = atoms.get_column(key)

        #     total_charge = configuration.charge
        #     n_atoms = configuration.n_atoms
        #     eex["charges"] = charges = [None] * n_atoms
        #     total_q = 0.0
        #     if "shell model" in terms:
        #         # Use charges from shell model by preference.
        #         shell_q = []
        #         for i, itype in enumerate(atom_types):
        #             tmp = self.shell_model(itype)
        #             if tmp is None:
        #                 # Fall back to charges
        #                 parameters = self.charges(itype)[3]
        #                 q = float(parameters["Q"])
        #                 charges[i] = q
        #                 total_q += q
        #             else:
        #                 parameters = tmp[3]
        #                 q = float(parameters["Q"])
        #                 y = float(parameters["Y"])
        #                 charges[i] = q - y
        #                 shell_q.append(y)
        #                 total_q += q
        #         charges.extend(shell_q)
        #     else:
        #         # First see if there are any templates
        #         molecule = configuration.to_RDKMol()
        #         fragments = self.get_fragments()
        #         for smiles, data in fragments.items():
        #             if "charges" not in data:
        #                 continue

        #             smarts = data["SMARTS"]
        #             pattern = rdkit.Chem.MolFromSmarts(smarts)
        #             matches = molecule.GetSubstructMatches(
        #                 pattern, maxMatches=6 * n_atoms
        #             )
        #             for match in matches:
        #                 for charge, atom in zip(data["charges"], match):
        #                     # Check if this has been assigned by another fragment
        #                     if charges[atom] is not None:
        #                         if charge != charges[atom]:
        #                             msg = (
        #                                 f"Error in fragment charges for atom {atom} "
        #                                 f"already has charge:\n{charges[atom]}\n"
        #                                 f"New assignment: {charge}"
        #                             )
        #                             logger.error(msg)
        #                             raise ForcefieldChargeError(msg)
        #                     charges[atom] = charge
        #         # Use the increments and charges for any left
        #         for i, itype in enumerate(atom_types):
        #             if charges[i] is None:
        #                 parameters = self.charges(itype)[3]
        #                 charges[i] = float(parameters["Q"])
        #             total_q += charges[i]
        #     if abs(total_q - total_charge) > 0.001:
        #         delta = (total_q - total_charge) / len(charges)
        #         charges = [q - delta for q in charges]
        #         logger.warning(
        #             f"The total charge from the forcefield, {total_q:3f}, does not "
        #             f"match the formal charge, {total_charge}."
        #             f"\nAdjusted each atom's charge by {-delta:.3f} to compensate."
        #         )
        #     logger.debug("Charges from charges:\n" + pprint.pformat(charges))
        # else:
        key = f"charges_{ff_name}"
        if key in atoms:
            eex["charges"] = [*atoms[key]]
        else:
            raise RuntimeError("No charges on system!")

        logger.debug("leaving eex_increment")

    def eex_atoms(self, eex, configuration):
        """List the atoms into the energy expression.

        Note that if using the shell model, an extra "atom" is added for the shell.
        At the moment the mass is split 90:10 between core and shell.
        The shells are appended at the end of the atoms so bonds, etc. work using
        the atom indices.
        """
        atoms = configuration.atoms
        n_atoms = configuration.n_atoms
        coordinates = atoms.get_coordinates(fractionals=False)
        if self.ff_form == "reaxff":
            types = atoms.symbols
        else:
            key = f"atom_types_{self.current_forcefield}"
            types = atoms.get_column(key)

        result = eex["atoms"] = []
        atom_types = eex["atom types"] = []
        masses = eex["masses"] = []
        elements = eex["elements"] = []

        shells = []
        eex["shell_of_atom"] = shell_of_atom = []

        for itype, xyz in zip(types, coordinates):
            x, y, z = xyz
            if self.shell_model(itype) is None:
                if itype in atom_types:
                    index = atom_types.index(itype) + 1
                else:
                    atom_types.append(itype)
                    index = len(atom_types)
                    masses.append((self.mass(itype), itype))
                    elements.append(self.element(itype))
                result.append((x, y, z, index))
                shell_of_atom.append(None)
            else:
                if itype in atom_types:
                    # core
                    index = atom_types.index("core_" + itype) + 1
                    result.append((x, y, z, index))
                    # shell
                    index = atom_types.index(itype) + 1
                    shell_of_atom.append(len(shells) + n_atoms)
                    shells.append((x, y, z, index))
                else:
                    # core
                    atom_types.append("core_" + itype)
                    index = len(atom_types)
                    masses.append((0.9 * float(self.mass(itype)), "core_" + itype))
                    elements.append(self.element(itype))
                    result.append((x, y, z, index))
                    # shell
                    atom_types.append(itype)
                    index = len(atom_types)
                    masses.append((0.1 * float(self.mass(itype)), "shell_" + itype))
                    shell_of_atom.append(len(shells) + n_atoms)
                    shells.append((x, y, z, index))

        if len(shells) > 0:
            result.extend(shells)

        eex["n_atoms"] = n_atoms = len(result)
        eex["n_atom_types"] = len(atom_types)

        # molecule for each atom and shell (if any)
        molecule = eex["molecule"] = [1] * n_atoms
        molecules = configuration.find_molecules(as_indices=True)
        for molecule_id, atoms in enumerate(molecules):
            for atom in atoms:
                molecule[atom] = molecule_id
                if shell_of_atom[atom] is not None:
                    molecule[shell_of_atom[atom]] = molecule_id

    def eex_pair(self, eex, configuration):
        """Create the pair (non-bond) portion of the energy expression"""
        logger.debug("In eex_pair")
        types = self.topology["types"]

        found = False
        for pair_type in ("nonbond(12-6)", "nonbond(9-6)", "buckingham"):
            if pair_type in self.ff["functional_forms"]:
                found = True
                break
        if not found:
            raise RuntimeError("Error finding pair_type in eex_pair")

        result = eex["nonbonds"] = []
        parameters = eex["nonbond parameters"] = []
        if pair_type == "buckingham":
            types = eex["atom types"]
            for i, itype in enumerate(types):
                if itype[0:5] == "core_":
                    itype = "core"
                for j, jtype in enumerate(types[0 : i + 1]):
                    if jtype[0:5] == "core_":
                        jtype = "core"
                    # print(f"{i}-{j} {itype} - {jtype} ##")
                    (
                        parameters_type,
                        real_types,
                        form,
                        parameter_values,
                    ) = self.nonbond_parameters(itype, jtype, form=pair_type)
                    new_value = (
                        form,
                        parameter_values,
                        (itype, jtype),
                        parameters_type,
                        real_types,
                    )
                    index = None
                    # print(f"{itype}-{jtype} --> {new_value}")
                    for value, count in zip(parameters, range(1, len(parameters) + 1)):
                        # print(f"\t{value}")
                        if self.eex_compare_values(value, new_value):
                            index = count
                            break
                    # if index is None:
                    parameters.append(new_value)
                    index = len(parameters)
                    # print(f"Added {new_value} as {index}")
                    result.append(index)
        else:
            for itype in types[1:]:
                if itype[0:5] == "core_":
                    itype = "core"
                (
                    parameters_type,
                    real_types,
                    form,
                    parameter_values,
                ) = self.nonbond_parameters(itype, form=pair_type)
                new_value = (
                    form,
                    parameter_values,
                    (itype,),
                    parameters_type,
                    real_types,
                )
                index = None
                for count, value in enumerate(parameters, start=1):
                    if value == new_value:
                        index = count
                        break
                if index is None:
                    parameters.append(new_value)
                    index = len(parameters)
                result.append(index)
        eex["n_nonbonds"] = len(result)
        eex["n_nonbond_types"] = len(parameters)

    def eex_shell_model(self, eex, configuration):
        """Create the shell model portion of the energy expression"""
        types = self.topology["types"]

        if "bonds" not in eex:
            eex["bonds"] = []
            eex["bond parameters"] = []
        result = eex["bonds"]
        parameters = eex["bond parameters"]

        key = f"atom_types_{self.current_forcefield}"
        types = configuration.atoms.get_column(key)
        shell_of_atom = eex["shell_of_atom"]
        n_atoms = configuration.n_atoms

        for atom_no, shell_no, itype in zip(range(n_atoms), shell_of_atom, types):
            if shell_no is not None:
                parameters_type, real_types, form, parameter_values = self.shell_model(
                    itype
                )
                real_type = real_types[0]
                new_value = (
                    form,
                    {"R0": 0.0, "K2": parameter_values["k"]},
                    ("core_" + itype, itype),
                    parameters_type,
                    ("core_" + real_type, real_type),
                )
                index = None
                for value, count in zip(parameters, range(1, len(parameters) + 1)):
                    if self.eex_compare_values(value, new_value):
                        index = count
                        break
                if index is None:
                    parameters.append(new_value)
                    index = len(parameters)
                result.append((atom_no + 1, shell_no + 1, index))

        eex["n_bonds"] = len(result)
        eex["n_bond_types"] = len(parameters)

    def eex_bond(self, eex, configuration):
        """Create the bond portion of the energy expression"""
        types = self.topology["types"]
        bonds = self.topology["bonds"]

        if "bonds" not in eex:
            eex["bonds"] = []
            eex["bond parameters"] = []
        result = eex["bonds"]
        parameters = eex["bond parameters"]
        for i, j in bonds:
            parameters_type, real_types, form, parameter_values = self.bond_parameters(
                types[i], types[j]
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j]),
                parameters_type,
                real_types,
            )
            index = None
            for value, count in zip(parameters, range(1, len(parameters) + 1)):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, index))
        eex["n_bonds"] = len(result)
        eex["n_bond_types"] = len(parameters)

    def eex_angle(self, eex, configuration):
        """Create the angle portion of the energy expression"""
        types = self.topology["types"]
        angles = self.topology["angles"]

        result = eex["angles"] = []
        parameters = eex["angle parameters"] = []
        for i, j, k in angles:
            parameters_type, real_types, form, parameter_values = self.angle_parameters(
                types[i], types[j], types[k]
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, index))
        eex["n_angles"] = len(result)
        eex["n_angle_types"] = len(parameters)

    def eex_torsion(self, eex, configuration):
        """Create the torsion portion of the energy expression"""
        types = self.topology["types"]
        torsions = self.topology["torsions"]

        result = eex["torsions"] = []
        parameters = eex["torsion parameters"] = []
        for i, j, k, l in torsions:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.torsion_parameters(types[i], types[j], types[k], types[l])
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for value, count in zip(parameters, range(1, len(parameters) + 1)):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_torsions"] = len(result)
        eex["n_torsion_types"] = len(parameters)

    def eex_out_of_plane(self, eex, configuration):
        """Create the out-of-plane portion of the energy expression"""
        types = self.topology["types"]
        oops = self.topology["oops"]

        result = eex["oops"] = []
        parameters = eex["oop parameters"] = []
        for i, j, k, l in oops:
            parameters_type, real_types, form, parameter_values = self.oop_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )

            if form == "dreiding_out_of_plane" and parameters_type == "zeroed":
                continue

            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_oops"] = len(result)
        eex["n_oop_types"] = len(parameters)

    def eex_bond_bond(self, eex, configuration):
        """Create the bond-bond portion of the energy expression"""
        types = self.topology["types"]
        angles = self.topology["angles"]

        result = eex["bond-bond"] = []
        parameters = eex["bond-bond parameters"] = []
        for i, j, k in angles:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.bond_bond_parameters(types[i], types[j], types[k], zero=True)
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k]),
                parameters_type,
                real_types,
            )
            # if self.eex_compare_values(value, new_value):
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, index))
        eex["n_bond-bond"] = len(result)
        eex["n_bond-bond_types"] = len(parameters)

    def eex_bond_angle(self, eex, configuration):
        """Create the bond-angle portion of the energy expression"""
        types = self.topology["types"]
        angles = self.topology["angles"]

        result = eex["bond-angle"] = []
        parameters = eex["bond-angle parameters"] = []
        for i, j, k in angles:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.bond_angle_parameters(types[i], types[j], types[k], zero=True)
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, index))
        eex["n_bond-angle"] = len(result)
        eex["n_bond-angle_types"] = len(parameters)

    def eex_torsion_middle_bond(self, eex, configuration):
        """Create the middle_bond-torsion portion of the energy expression"""
        types = self.topology["types"]
        torsions = self.topology["torsions"]

        result = eex["middle_bond-torsion_3"] = []
        parameters = eex["middle_bond-torsion_3 parameters"] = []
        for i, j, k, l in torsions:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.middle_bond_torsion_3_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_middle_bond-torsion_3"] = len(result)
        eex["n_middle_bond-torsion_3_types"] = len(parameters)

    def eex_torsion_end_bond(self, eex, configuration):
        """Create the end_bond-torsion portion of the energy expression"""
        types = self.topology["types"]
        torsions = self.topology["torsions"]

        result = eex["end_bond-torsion_3"] = []
        parameters = eex["end_bond-torsion_3 parameters"] = []
        for i, j, k, l in torsions:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.end_bond_torsion_3_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_end_bond-torsion_3"] = len(result)
        eex["n_end_bond-torsion_3_types"] = len(parameters)

    def eex_torsion_angle(self, eex, configuration):
        """Create the angle-torsion portion of the energy expression"""
        types = self.topology["types"]
        torsions = self.topology["torsions"]

        result = eex["angle-torsion_3"] = []
        parameters = eex["angle-torsion_3 parameters"] = []
        for i, j, k, l in torsions:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.angle_torsion_3_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_angle-torsion_3"] = len(result)
        eex["n_angle-torsion_3_types"] = len(parameters)

    def eex_angle_torsion_angle(self, eex, configuration):
        """Create the angle-angle-torsion portion of the energy expression"""
        types = self.topology["types"]
        torsions = self.topology["torsions"]

        result = eex["angle-angle-torsion_1"] = []
        parameters = eex["angle-angle-torsion_1 parameters"] = []
        for i, j, k, l in torsions:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.angle_angle_torsion_1_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_angle-angle-torsion_1"] = len(result)
        eex["n_angle-angle-torsion_1_types"] = len(parameters)

    def eex_1_3_bond_bond(self, eex, configuration):
        """Create the 1,3 bond-bond portion of the energy expression"""
        types = self.topology["types"]
        torsions = self.topology["torsions"]

        result = eex["bond-bond_1_3"] = []
        parameters = eex["bond-bond_1_3 parameters"] = []
        for i, j, k, l in torsions:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.bond_bond_1_3_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )
            new_value = (
                form,
                parameter_values,
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for count, value in enumerate(parameters, start=1):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_bond-bond_1_3"] = len(result)
        eex["n_bond-bond_1_3_types"] = len(parameters)

    def eex_angle_angle(self, eex, configuration):
        """Create the angle-angle portion of the energy expression

        j is the vertex atom of the angles. For the angle-angle parameters
        the bond j-k is the common bond, i.e. the angles are i-j-k and j-k-l
        """
        types = self.topology["types"]
        oops = self.topology["oops"]

        result = eex["angle-angle"] = []
        parameters = eex["angle-angle parameters"] = []
        for i, j, k, l in oops:
            (
                parameters_type,
                real_types,
                form,
                parameter_values,
            ) = self.angle_angle_parameters(
                types[i], types[j], types[k], types[l], zero=True
            )
            K1 = parameter_values["K"]
            Theta10 = parameter_values["Theta10"]
            Theta30 = parameter_values["Theta20"]
            oK1 = parameter_values["original K"]
            oTheta10 = parameter_values["original Theta10"]
            oTheta30 = parameter_values["original Theta20"]
            tmp = self.angle_angle_parameters(
                types[k], types[j], types[i], types[l], zero=True
            )[3]
            K2 = tmp["K"]
            Theta20 = tmp["Theta20"]
            oK2 = tmp["original K"]
            oTheta20 = tmp["original Theta20"]
            tmp = self.angle_angle_parameters(
                types[i], types[j], types[l], types[k], zero=True
            )[3]
            K3 = tmp["K"]
            oK3 = tmp["original K"]
            new_value = (
                form,
                {
                    "K1": K1,
                    "K2": K2,
                    "K3": K3,
                    "Theta10": Theta10,
                    "Theta20": Theta20,
                    "Theta30": Theta30,
                    "original K1": oK1,
                    "original K2": oK2,
                    "original K3": oK3,
                    "original Theta10": oTheta10,
                    "original Theta20": oTheta20,
                    "original Theta30": oTheta30,
                },
                (types[i], types[j], types[k], types[l]),
                parameters_type,
                real_types,
            )
            index = None
            for value, count in zip(parameters, range(1, len(parameters) + 1)):
                if self.eex_compare_values(value, new_value):
                    index = count
                    break
            if index is None:
                parameters.append(new_value)
                index = len(parameters)
            result.append((i, j, k, l, index))
        eex["n_angle-angle"] = len(result)
        eex["n_angle-angle_types"] = len(parameters)

    def eex_compare_values(self, old, new):
        """Compare parameters values to see if they are the same."""
        if self.ff_form == "class2":
            # Lammps class2 relies on having all the angles, torsions, and oops
            # for the cross terms to match.
            return old == new

        return old[0] == new[0] and old[1] == new[1] and old[4] == new[4]



# -*- coding: utf-8 -*-

"""Main class for handling forcefields"""

import logging

import rdkit
import rdkit.Chem
import rdkit.Chem.Draw
import rdkit.Chem.AllChem

logger = logging.getLogger(__name__)
# logger.setLevel("DEBUG")


class ForcefieldAssignmentError(Exception):
    """The forcefield could not be assigned successfully."""

    def __init__(self, message="The atom-types could not be assigned."):
        super().__init__(message)


class FFAssigner(object):
    def __init__(self, forcefield):
        """Handle the assignment of the forcefield to the structure

        This class is closely related to the Forcefield class, but
        separated from it due to the dependencies it carries along,
        coupled with the fact that it is not needed in some
        computations where the forcefield itself is.
        """

        self.forcefield = forcefield

    def assign(self, configuration, picture=False):
        """Assign the atom types to the structure using SMARTS templates"""
        if self.forcefield.ff_form in ("reaxff",):
            # Nothing to do.
            return

        molecule = configuration.to_RDKMol()

        n_atoms = configuration.n_atoms

        atom_types = ["?"] * n_atoms

        # First try the fragments, if any
        fragments = self.forcefield.get_fragments()
        for smiles, data in fragments.items():
            smarts = data["SMARTS"]

            pattern = rdkit.Chem.MolFromSmarts(smarts)

            matches = molecule.GetSubstructMatches(pattern, maxMatches=6 * n_atoms)
            logger.debug(smiles + ": ")
            if len(matches) > 0:
                for match in matches:
                    for _type, atom in zip(data["atom types"], match):
                        # Check if this has already been assigned to another fragment
                        if atom_types[atom] != "?":
                            key1 = atom_types[atom].split("_")[0]
                            key2 = _type.split("_")[0]
                            if key2 != key1:
                                self.draw_atom_types(
                                    molecule,
                                    atom_types,
                                    filename="duplicate_atom_types.png",
                                )
                                msg = (
                                    f"Error in fragment atom typing: atom {atom} "
                                    f"already typed:\n{atom_types[atom]}\n"
                                    f"New assignment: {_type}"
                                )
                                logger.error(msg)
                                raise ForcefieldAssignmentError(msg)
                        atom_types[atom] = _type

        # Fragments have priority, so don't change their assignments
        assigned = [type_ != "?" for type_ in atom_types]

        # And now the templates for the atom_types
        templates = self.forcefield.get_templates()
        for atom_type in templates:
            template = templates[atom_type]
            for smarts in template["smarts"]:
                pattern = rdkit.Chem.MolFromSmarts(smarts)

                ind_map = {}
                for atom in pattern.GetAtoms():
                    map_num = atom.GetAtomMapNum()
                    if map_num:
                        ind_map[map_num - 1] = atom.GetIdx()
                map_list = [ind_map[x] for x in sorted(ind_map)]

                matches = molecule.GetSubstructMatches(pattern, maxMatches=6 * n_atoms)
                logger.debug(atom_type + ": ")
                if len(matches) > 0:
                    for match in matches:
                        atom_ids = [match[x] for x in map_list]
                        for x in atom_ids:
                            if not assigned[x]:
                                atom_types[x] = atom_type
                        tmp = [str(x) for x in atom_ids]
                        logger.debug("\t" + ", ".join(tmp))

        # Create a picture with the atom types if requested
        if picture:
            self.draw_atom_types(molecule, atom_types)

        i = 0
        untyped = []
        for atom, atom_type in zip(molecule.GetAtoms(), atom_types):
            if atom_type == "?":
                untyped.append(i)
            logger.debug("{}: {}".format(atom.GetSymbol(), atom_type))
            i += 1

        if len(untyped) > 0:
            msg = (
                "The forcefield does not have atom types for"
                " the molecule!. See missing_atom_types.png"
                " for more detail."
            )
            logger.error(msg)
            self.draw_atom_types(
                molecule,
                atom_types,
                filename="missing_atom_types.png",
                highlight_atoms=untyped,
            )
            raise ForcefieldAssignmentError(msg)
        else:
            logger.info("The molecule was successfully atom-typed")

        return atom_types

    def draw_atom_types(
        self,
        molecule,
        atom_types,
        legend=None,
        filename="atom_types.png",
        highlight_atoms=None,
    ):
        """Create a picture of the molecule labeled by atom types.

        Parameters
        ----------
        molecule : rdkit.molecule
            The molecule as an RDKit molecule.
        atom_types : [str]
            The labels for the atom types
        legend : str
            The legend for the picture. Defaults to "Atom types for ff_file:ffname"
        filename : str or pathlib.Path
            The file to write the picture to.
        highlight_atoms : [int]
            The atoms to highlight
        """
        if legend is None:
            legend = (
                f"Atom Types for {self.forcefield.filename}:"
                f"{self.forcefield.current_forcefield}"
            )

        for atom, atom_type in zip(molecule.GetAtoms(), atom_types):
            atom.SetProp("atomNote", atom_type)
        rdkit.Chem.AllChem.Compute2DCoords(molecule)
        rdkit.Chem.rdDepictor.StraightenDepiction(molecule)

        d2d = rdkit.Chem.Draw.MolDraw2DCairo(1000, 1000)
        dopts = d2d.drawOptions()
        dopts.addAtomIndices = True
        dopts.legendFontSize = 30

        d2d.DrawMolecule(molecule, legend=legend, highlightAtoms=highlight_atoms)
        d2d.FinishDrawing()
        d2d.WriteDrawingText(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `seamm_ff_util` package."""

import json
import seamm_ff_util  # noqa: F401


def test_end_bond_torsion_3_explicit(pcff):
    """Test of end_bond_torsion_3 parameters, which should find
    explicit ones"""

    expected = {
        "reference": "1",
        "version": "1.0",
        "V1_L": "0.0870",
        "V2_L": "0.5143",
        "V3_L": "-0.2448",
        "V1_R": "0.2217",
        "V2_R": "0.4780",
        "V3_R": "-0.0817",
        "R0_L": "1.5140",
        "R0_R": "1.1010",
        "original V1_L": "0.0870",
        "original V2_L": "0.5143",
        "original V3_L": "-0.2448",
        "original V1_R": "0.2217",
        "original V2_R": "0.4780",
        "original V3_R": "-0.0817",
        "original R0_L": "1.5140",
        "original R0_R": "1.1010",
    }

    i = "h"
    j = "c"
    k = "c"
    l = "c_0"  # noqa: E741
    ptype, key, form, parameters = pcff.end_bond_torsion_3_parameters(i, j, k, l)
    assert ptype == "explicit"
    assert key == ("h", "c", "c", "c_0")
    if parameters != expected:
        print(json.dumps(parameters, indent=4))
    assert parameters == expected


def test_end_bond_torsion_3_explicit_kji(pcff):
    """known end_bond_torsion_3 parameters, ordered backwards"""

    expected = {
        "reference": "1",
        "version": "1.0",
        "V1_L": "0.2217",
        "original V1_L": "0.2217",
        "V2_L": "0.4780",
        "original V2_L": "0.4780",
        "V3_L": "-0.0817",
        "original V3_L": "-0.0817",
        "V1_R": "0.0870",
        "original V1_R": "0.0870",
        "V2_R": "0.5143",
        "original V2_R": "0.5143",
        "V3_R": "-0.2448",
        "original V3_R": "-0.2448",
        "R0_L": "1.5140",
        "R0_R": "1.1010",
        "original R0_L": "1.5140",
        "original R0_R": "1.1010",
    }

    i = "c_0"
    j = "c"
    k = "c"
    l = "h"  # noqa: E741
    ptype, key, form, parameters = pcff.end_bond_torsion_3_parameters(i, j, k, l)
    assert ptype == "explicit"
    assert key == ("c_0", "c", "c", "h")
    if parameters != expected:
        print(json.dumps(parameters, indent=4))
    assert parameters == expected


def test_end_bond_torsion_3_equivalent(pcff):
    """Simple test of end_bond_torsion_3 parameters using equivalencies"""
    expected = {
        "reference": "1",
        "version": "1.0",
        "V1_L": "1.3997",
        "original V1_L": "1.3997",
        "V2_L": "0.7756",
        "original V2_L": "0.7756",
        "V3_L": "0.0000",
        "original V3_L": "0.0000",
        "V1_R": "-0.5835",
        "original V1_R": "-0.5835",
        "V2_R": "1.1220",
        "original V2_R": "1.1220",
        "V3_R": "0.3978",
        "original V3_R": "0.3978",
        "R0_L": "1.1010",
        "R0_R": "1.4170",
        "original R0_L": "1.1010",
        "original R0_R": "1.4170",
    }

    i = "h"
    j = "c"
    k = "c5"
    l = "c5"  # noqa: E741
    ptype, key, form, parameters = pcff.end_bond_torsion_3_parameters(i, j, k, l)
    assert ptype == "equivalent"
    assert key == ("h", "c", "cp", "cp")
    if parameters != expected:
        print(json.dumps(parameters, indent=4))
    assert parameters == expected
